/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include <cassert>
#include <cstring>
#include <memory>
#include <string>
#include <vector>
#include "defs.h"
#include "record/rm_defs.h"
#include <parser/ast.h>


struct TabCol {
    std::string tab_name;
    std::string col_name;

    friend bool operator<(const TabCol &x, const TabCol &y) {
        return std::make_pair(x.tab_name, x.col_name) < std::make_pair(y.tab_name, y.col_name);
    }

    TabCol& operator=(const TabCol& other) {
        if (this != &other) {
            tab_name = other.tab_name;
            col_name = other.col_name;
        }
        return *this;
    }
};

// 自定义哈希函数
struct TabColHash {
    size_t operator()(const TabCol& tc) const {
        std::hash<std::string> string_hash;
        return string_hash(tc.tab_name) ^ string_hash(tc.col_name);
    }
};

// 自定义等价比较函数
struct TabColEqual {
    bool operator()(const TabCol& lhs, const TabCol& rhs) const {
        return std::tie(lhs.tab_name, lhs.col_name) == std::tie(rhs.tab_name, rhs.col_name);
    }
};

struct Value {
    ColType type;  // type of value
    union {
        int int_val;      // int value
        float float_val;  // float value
    };
    std::string str_val;  // string value

    std::shared_ptr<RmRecord> raw;  // raw record buffer

    void set_int(int int_val_) {
        type = TYPE_INT;
        int_val = int_val_;
    }

    void set_float(float float_val_) {
        type = TYPE_FLOAT;
        float_val = float_val_;
    }

    void set_str(std::string str_val_) {
        type = TYPE_STRING;
        str_val = std::move(str_val_);
    }

    void init_raw(int len) {
        assert(raw == nullptr);
        raw = std::make_shared<RmRecord>(len);
        if (type == TYPE_INT) {
            assert(len == sizeof(int));
            *(int *)(raw->data) = int_val;
        } else if (type == TYPE_FLOAT) {
            assert(len == sizeof(float));
            *(float *)(raw->data) = float_val;
        } else if (type == TYPE_STRING) {
            if (len < (int)str_val.size()) {
                throw StringOverflowError();
            }
            memset(raw->data, 0, len);
            memcpy(raw->data, str_val.c_str(), str_val.size());
        }
    }
        
    void str_to_value(std::string str, ColType valtype) {
        if(valtype == ColType::TYPE_INT)
        {
            set_int(std::stoi(str));
        }else if(valtype == ColType::TYPE_FLOAT)
        {
            set_float(std::stof(str));  // 可能会有精度问题 lun
        }else if(valtype == ColType::TYPE_STRING)
        {
            set_str(str);
        }
    }
};

enum CompOp { OP_EQ, OP_NE, OP_LT, OP_GT, OP_LE, OP_GE };

struct Condition {
    TabCol lhs_col;   // left-hand side column
    CompOp op;        // comparison operator
    bool is_rhs_val;  // true if right-hand side is a value (not a column)
    TabCol rhs_col;   // right-hand side column
    Value rhs_val;    // right-hand side value
};

struct SetClause {
    TabCol lhs;
    Value rhs;
    bool is_add =false;
};

struct rcd_segment{
    char *data;
    int len;
    int point;
};

struct AggregateFunction
{
    ast::AggregateType  type;
    std::string as_col;
    TabCol agg_col;
    
    AggregateFunction(ast::AggregateType type_,std::string as_col_,TabCol agg_col_):
    type(type_),as_col(as_col_),agg_col(agg_col_)
    {};
    AggregateFunction& operator=(const AggregateFunction& other) {
        if (this != &other) {
            type = other.type;
            as_col = std::move(other.as_col);
            agg_col = std::move(other.agg_col);
        }
        return *this;
    };
};

struct Agg_Cols
{
    bool is_agg;
    ast::AggregateType  type;
    TabCol col;
    std::string as_name;

    Agg_Cols(bool is_agg_,TabCol col_,ast::AggregateType type_,std::string as_name_):
    is_agg(is_agg_),col(col_),type(type_),as_name(as_name_){};
};

struct groupkey{
    int key_len;
    char *key_data;
    groupkey(int key_len_,char * key_data_){
        key_len=key_len_;
        key_data=key_data_;
    }
};

//自定义返回值相等的哈希函数，便于使用unordered_map
struct CustomHash {
    std::size_t operator()(const groupkey* key) const {
        return 0;
    }
};

//自定义一个比较函数，为了便于分组比较
struct CustomEqual {
    bool operator()(const groupkey* lhs, const groupkey* rhs) const {
        return std::memcmp(lhs->key_data, rhs->key_data,lhs->key_len) == 0;
    }
};

struct HavingCondition {
    AggregateFunction lhs_col;   // left-hand side column
    CompOp op;        // comparison operator
    bool is_rhs_val;  // true if right-hand side is a value (not a column)
    TabCol rhs_col;   // right-hand side column
    Value rhs_val;    // right-hand side value

    HavingCondition(AggregateFunction lhs_col_,CompOp op_,bool is_rhs_val_,TabCol rhs_col_,Value rhs_val_):
    lhs_col(lhs_col_),op(op_),is_rhs_val(is_rhs_val_),rhs_col(rhs_col_),rhs_val(rhs_val_)
    {}
};

struct  Having
{
    std::vector<Condition> conds;
    std::vector<HavingCondition> havingconds;
};



