/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class IndexScanExecutor : public AbstractExecutor {
   private:
    std::string tab_name_;                      // 表名称
    TabMeta tab_;                               // 表的元数据
    
    RmFileHandle *fh_;                          // 表的数据文件句柄
    std::vector<ColMeta> cols_;                 // 需要读取的字段
    size_t len_;                                // 选取出来的一条记录的长度
    std::vector<Condition> fed_conds_;          // 扫描条件，和conds_字段相同

    std::vector<std::string> index_col_names_;  // index scan涉及到的索引包含的字段
    IndexMeta index_meta_;                      // index scan涉及到的索引元数据

    Rid rid_;
    std::unique_ptr<RecScan> scan_;

    SmManager *sm_manager_;

   public:
    std::vector<Condition> conds_;              // 扫描条件

    IndexScanExecutor(SmManager *sm_manager, std::string tab_name, std::vector<Condition> conds, std::vector<std::string> index_col_names,
                    Context *context, bool flag) {
        sm_manager_ = sm_manager;
        context_ = context;
        tab_name_ = std::move(tab_name);
        tab_ = sm_manager_->db_.get_table(tab_name_);
        conds_ = std::move(conds);
        // index_no_ = index_no;
        index_col_names_ = index_col_names; 
        index_meta_ = *(tab_.get_index_meta(index_col_names_));
        fh_ = sm_manager_->fhs_.at(tab_name_).get();
        cols_ = tab_.cols;
        len_ = cols_.back().offset + cols_.back().len;
        std::map<CompOp, CompOp> swap_op = {
            {OP_EQ, OP_EQ}, {OP_NE, OP_NE}, {OP_LT, OP_GT}, {OP_GT, OP_LT}, {OP_LE, OP_GE}, {OP_GE, OP_LE},
        };

        for (auto &cond : conds_) {
            if (cond.lhs_col.tab_name != tab_name_) {
                // lhs is on other table, now rhs must be on this table
                assert(!cond.is_rhs_val && cond.rhs_col.tab_name == tab_name_);
                // swap lhs and rhs
                std::swap(cond.lhs_col, cond.rhs_col);
                cond.op = swap_op.at(cond.op);
            }
        }
        fed_conds_ = conds_;

        has_index = flag;
    }

    //函数的作用主要是初始化索引扫描的算子，给出初始化所需的上下界
    void beginTuple() override {
        //上读锁
        if(context_!=nullptr){
            // context_->lock_mgr_->lock_shared_on_table(context_->txn_, fh_->GetFd());
            context_->lock_mgr_->lock_IS_on_table(context_->txn_, fh_->GetFd());
        }
        // context_->lock_mgr_->lock_shared_on_table(context_->txn_, fh_->GetFd());
        //初始化上界和下界
        RmRecord lower_bound(index_meta_.col_tot_len),upper_bound(index_meta_.col_tot_len);
        //初始化上下界
        int offset=0;
        for(int i=0;i<index_meta_.col_num;i++){
            Value col_lower,col_upper;
            if(index_meta_.cols[i].type==TYPE_INT){
                col_lower.set_int(INT32_MIN);
                col_upper.set_int(INT32_MAX);
            }else if(index_meta_.cols[i].type==TYPE_FLOAT){
                col_lower.set_float(-__FLT_MAX__);
                col_upper.set_float(__FLT_MAX__);
            }else if(index_meta_.cols[i].type==TYPE_STRING){
                std::string max_str=std::string(index_meta_.cols[i].len,255);
                std::string min_str=std::string(index_meta_.cols[i].len,0);
                col_lower.set_str(min_str);
                col_upper.set_str(max_str);
            }
            col_lower.init_raw(index_meta_.cols[i].len);
            col_upper.init_raw(index_meta_.cols[i].len);
            //根据条件修改上下界
            for(auto cond:conds_){
                if(index_meta_.cols[i].name==cond.lhs_col.col_name&&cond.is_rhs_val){
                    if(cond.op==OP_EQ){
                        col_upper=cond.rhs_val;
                        col_lower=cond.rhs_val;
                    }else if(cond.op==OP_GE){
                        if(ix_compare(cond.rhs_val.raw->data,
                                    col_lower.raw->data,
                                    index_meta_.cols[i].type,
                                    index_meta_.cols[i].len)==1){
                            col_lower=cond.rhs_val;
                        }
                    }else if(cond.op==OP_GT){
                        if(ix_compare(cond.rhs_val.raw->data,
                                    col_lower.raw->data,
                                    index_meta_.cols[i].type,
                                    index_meta_.cols[i].len)==1){
                            col_lower=cond.rhs_val;
                        }
                    }else if(cond.op==OP_LE){
                        if(ix_compare(cond.rhs_val.raw->data,
                                    col_upper.raw->data,
                                    index_meta_.cols[i].type,
                                    index_meta_.cols[i].len)==-1){
                            col_upper=cond.rhs_val;
                        }
                    }else if(cond.op==OP_LT){
                        if(ix_compare(cond.rhs_val.raw->data,
                                    col_upper.raw->data,
                                    index_meta_.cols[i].type,
                                    index_meta_.cols[i].len)==-1){
                            col_upper=cond.rhs_val;
                        }
                    }else if(cond.op==OP_NE){
                        //全表扫
                    }
                }
            }
            memcpy(upper_bound.data+offset,col_upper.raw->data,index_meta_.cols[i].len);
            memcpy(lower_bound.data+offset,col_lower.raw->data,index_meta_.cols[i].len);
            offset=offset+index_meta_.cols[i].len;
        }
        //上间隙锁
        if(context_!=nullptr){
            context_->lock_mgr_->lock_GAP_on_table(context_->txn_, fh_->GetFd(),&index_meta_,upper_bound.data,lower_bound.data,offset,tab_name_);
        }
        //根据上下界的data找到index的上下界，初始化indexscan
        auto ih=sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_,index_meta_.cols)).get();
        Iid upper_index=ih->upper_bound(upper_bound.data);
        Iid lower_index=ih->lower_bound(lower_bound.data);
        scan_=std::make_unique<IxScan>(ih,lower_index,upper_index,sm_manager_->get_bpm());
        //根据cond过滤
        while(!scan_->is_end()){
            rid_=scan_->rid();
            bool success=true;
            // RmRecord* rec;
            std::unique_ptr<RmRecord> t;
            try{
                t=fh_->get_record(rid_,context_);
                // rec=(fh_->get_record(rid_,context_)).get();
            }catch(RecordNotFoundError& e){
                std::cout<<"record not found"<<std::endl;
                //没有拿到记录判断为false,这部分的异常处理没有想好
                success=false;
            }
            if(success){
                bool is_equal=isequal(cols_,conds_,t);
                if(is_equal){
                    break;
                }
            }
            scan_->next();
        }
    }
    bool isequal(const std::vector<ColMeta>& tab_cols,const std::vector<Condition>& conds,const std::unique_ptr<RmRecord>& rec){
        //循环比较一个记录的内容与每种where条件
        bool is_equal=true;
        for(const auto& cond:conds){
            //取出每条记录的列
            auto lhs_col=get_col(tab_cols,cond.lhs_col);
            //取出对应列的data
            char* lhs_data=rec->data+lhs_col->offset;
            ColType lhs_type=lhs_col->type;
            char* rhs_data;
            //取出右值
            if(cond.is_rhs_val){
                //右值把值存在记录里面了
                rhs_data=cond.rhs_val.raw->data;
            }else{
                auto rhs_col=get_col(tab_cols,cond.rhs_col);
                rhs_data=rec->data+rhs_col->offset;
                
            }
            
            //使用ix_compare函数进行对比
            const int col_len=lhs_col->len;
            int num=ix_compare(lhs_data,rhs_data,lhs_type,col_len);
            if(cond.op==OP_EQ){
                is_equal=(num==0)&&is_equal;
            }else if(cond.op==OP_GE){
                is_equal=(num>=0)&&is_equal;
            }else if(cond.op==OP_GT){
                is_equal=(num>0)&&is_equal;
            }else if(cond.op==OP_LE){
                is_equal=(num<=0)&&is_equal;
            }else if(cond.op==OP_LT){
                is_equal=(num<0)&&is_equal;
            }else{
                is_equal=(num!=0)&&is_equal;
            }
            if(is_equal==false){
                return is_equal;
            }
        }
        return is_equal;
    }
    void nextTuple() override {
        scan_->next();
         while(!scan_->is_end()){
            rid_=scan_->rid();
            bool success=true;
            // RmRecord* rec;
            std::unique_ptr<RmRecord> t;
            try{
                t=fh_->get_record(rid_,context_);
                // rec=(fh_->get_record(rid_,context_)).get();
            }catch(RecordNotFoundError& e){
                std::cout<<"record not found"<<std::endl;
                //没有拿到记录判断为false,这部分的异常处理没有想好
                success=false;
            }
            if(success){
                bool is_equal=isequal(cols_,conds_,t);
                if(is_equal){
                    break;
                }
            }
            scan_->next();
        }
    }

    std::unique_ptr<RmRecord> Next() override {
        return fh_->get_record(rid_,context_);
    }

    Rid &rid() override { return rid_; }
    // Rid &rid() override { return _abstract_rid; }

    std::string getType() override { return "indexscan"; };

    bool is_end() const override { return scan_->is_end(); };

    ColMeta get_col_offset(const TabCol &target){
        for(int i=0;i<cols_.size();i++){
            if(cols_[i].name==target.col_name&&cols_[i].tab_name==target.tab_name){
                return cols_[i];
            }
        }
    }

    const std::vector<ColMeta> &cols() const override {
        return cols_;
    }

    size_t tupleLen() const override{ return len_; };

    std::string get_table_name(){
        return tab_name_;
    }
};