/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include <mutex>
#include <condition_variable>
#include "transaction/transaction.h"
#include"record/rm_file_handle.h"
// #include "system/sm.h"
// #include"system/sm_meta.h"

static const std::string GroupLockModeStr[10] = {"NON_LOCK", "IS", "IX", "S", "X", "SIX"};

class SmManager;

class LockManager {
    /* 加锁类型，包括共享锁、排他锁、意向共享锁、意向排他锁、SIX（意向排他锁+共享锁） */
    enum class LockMode { INTENTION_SHARED,INTENTION_EXCLUSIVE,SHARED,S_IX , EXLUCSIVE};

    /* 用于标识加锁队列中排他性最强的锁类型，例如加锁队列中有SHARED和EXLUSIVE两个加锁操作，则该队列的锁模式为X */
    enum class GroupLockMode { NON_LOCK, IS, IX, S, X, SIX};

    int COMPATIBLITYMATRIX[5][5] = {
        {1, 1, 1, 1,0},
        {1, 1, 0, 0,0},
        {1, 0, 1, 0,0},
        {1, 0, 0, 0,0},
        {0, 0, 0, 0,0}
    };

    /* 事务的加锁申请 */
    class LockRequest {
    public:
        LockRequest(txn_id_t txn_id, LockMode lock_mode,bool granted,timestamp_t txn_start_ts_)
            : txn_id_(txn_id), lock_mode_(lock_mode), granted_(granted),txn_start_ts(txn_start_ts_) {}

        txn_id_t txn_id_;   // 申请加锁的事务ID
        LockMode lock_mode_;    // 事务申请加锁的类型
        bool granted_;          // 该事务是否已经被赋予锁
        timestamp_t txn_start_ts;//事务时间戳
    };

    /* 数据项上的加锁队列 */
    class LockRequestQueue {
    public:
        std::list<std::shared_ptr<LockRequest>> request_queue_;  // 加锁队列
        std::condition_variable cv_;            // 条件变量，用于唤醒正在等待加锁的申请，在no-wait策略下无需使用
        GroupLockMode group_lock_mode_ = GroupLockMode::NON_LOCK;   // 加锁队列的锁模式
    };

public:
    LockManager(SmManager *sm_manager):sm_manager_(sm_manager) {}

    ~LockManager() {}

    bool isCompatible(LockMode existingLock, LockMode requestedLock);

    bool lock_shared_on_record(Transaction* txn, const Rid& rid, int tab_fd);

    bool lock_exclusive_on_record(Transaction* txn, const Rid& rid, int tab_fd,char* rec);

    bool lock_shared_on_table(Transaction* txn, int tab_fd);

    bool lock_exclusive_on_table(Transaction* txn, int tab_fd);

    bool lock_IS_on_table(Transaction* txn, int tab_fd);

    bool lock_IX_on_table(Transaction* txn, int tab_fd);

    bool unlock(Transaction* txn, LockDataId lock_data_id);

    bool lock_GAP_on_table(Transaction* txn, int tab_fd,IndexMeta* index_meta,char* upper,char* lower,int len,std::string tab_name);

    bool general_lock(Transaction* txn, LockDataId lock_data_id,LockMode lock_mode);

    bool check(Transaction *txn);


    bool can_getlock(Transaction* txn, LockDataId lock_data_id,LockMode lock_mode){
        switch (lock_table_[lock_data_id].group_lock_mode_)
        {
        case GroupLockMode::NON_LOCK:
            lock_table_[lock_data_id].group_lock_mode_ = change_grouplockmode(lock_mode);
            return true;
        case GroupLockMode::IS:
            if (lock_mode != LockMode::EXLUCSIVE){
                lock_table_[lock_data_id].group_lock_mode_ = change_grouplockmode(lock_mode);
                return true;
            }
            break;
        case GroupLockMode::IX:
            if (lock_mode == LockMode::INTENTION_EXCLUSIVE || lock_mode == LockMode::INTENTION_SHARED)
                return true;
            break;
        case GroupLockMode::S:
            if (lock_mode == LockMode::SHARED || lock_mode == LockMode::INTENTION_SHARED)
                return true;
            break;
        case GroupLockMode::SIX:
            if (lock_mode == LockMode::INTENTION_SHARED)
                return true;
            break;
        case GroupLockMode::X:
            return false;
        }
        return false;
    }

    GroupLockMode change_grouplockmode(LockMode lock_mode){
        GroupLockMode ret = GroupLockMode::NON_LOCK;
        if(lock_mode==LockMode::SHARED){
            ret = GroupLockMode::S;
        }else if(lock_mode==LockMode::EXLUCSIVE){
            ret = GroupLockMode::X;
        }else if(lock_mode==LockMode::INTENTION_SHARED){
            ret = GroupLockMode::IS;
        }else if(lock_mode==LockMode::INTENTION_EXCLUSIVE){
            ret = GroupLockMode::IX;
        }else if(LockMode::S_IX==lock_mode){
            ret = GroupLockMode::SIX;
        }
        return ret;
    }

    LockDataId* is_gap_compatible(LockDataId lock_data_id,char* inrec_data,Transaction* txn);
    SmManager *sm_manager_;
private:
    std::mutex latch_;      // 用于锁表的并发
    std::unordered_map<LockDataId, LockRequestQueue> lock_table_;   // 全局锁表
    std::unordered_map<LockDataId, std::unique_ptr<LockRequestQueue>> Gap_lock_table_;//间隙锁表
    
};
