/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "transaction_manager.h"
#include "record/rm_file_handle.h"
#include "system/sm_manager.h"

std::unordered_map<txn_id_t, Transaction *> TransactionManager::txn_map = {};

/**
 * @description: 事务的开始方法
 * @return {Transaction*} 开始事务的指针
 * @param {Transaction*} txn 事务指针，空指针代表需要创建新事务，否则开始已有事务
 * @param {LogManager*} log_manager 日志管理器指针
 */
Transaction * TransactionManager::begin(Transaction* txn, LogManager* log_manager) {
    // Todo:
    // 1. 判断传入事务参数是否为空指针
    //保护全局map
    std::scoped_lock lock{latch_};
    // 2. 如果为空指针，创建新事务
    if(txn==nullptr){
        txn=new Transaction(next_txn_id_,IsolationLevel::SERIALIZABLE);
        next_txn_id_++;
    }
    // 3. 把开始事务加入到全局事务表中
    txn_map[txn->get_transaction_id()]=txn;
    //写进begin日志里面
    BeginLogRecord* begin_log=new BeginLogRecord(txn->get_transaction_id());
    //先写进日志缓冲区？
    log_manager->add_log_to_buffer(begin_log);
    log_manager->flush_log_to_disk();
    //设置事务的上一条lsn
    txn->set_prev_lsn(begin_log->lsn_);
    //设置事务时间戳
    txn->set_start_ts(next_timestamp_);
    next_timestamp_++;
    delete begin_log;
    //???设置事务状态为开始
    txn->set_state(TransactionState::DEFAULT);
    // 4. 返回当前事务指针

    
    return txn;
}

/**
 * @description: 事务的提交方法
 * @param {Transaction*} txn 需要提交的事务
 * @param {LogManager*} log_manager 日志管理器指针
 */
void TransactionManager::commit(Transaction* txn, LogManager* log_manager) {
    // Todo:
    // 1. 如果存在未提交的写操作，提交所有的写操作
    std::scoped_lock lock{latch_};
    CommitLogRecord* commit_log=new CommitLogRecord(txn->get_transaction_id());
    commit_log->prev_lsn_=txn->get_prev_lsn();
    log_manager->add_log_to_buffer(commit_log);
    log_manager->flush_log_to_disk();
    txn->set_prev_lsn(commit_log->lsn_);
    if(txn->get_write_set()->size()>0){
        auto txn_write_set=txn->get_write_set();
        while(txn_write_set->size()>0){
            auto write=txn_write_set->back();
            txn_write_set->pop_back();
            delete write;
        }
        txn_write_set->clear();
    }
    // 2. 释放所有锁
    for(auto &lock:*(txn->get_lock_set())){
        lock_manager_->unlock(txn,lock);
    }
    // 3. 释放事务相关资源，eg.锁集
    txn->get_lock_set()->clear();
    // 4. 把事务日志刷入磁盘中
    
    // 5. 更新事务状态
    txn->set_state(TransactionState::COMMITTED);
    delete commit_log;
}

/**
 * @description: 事务的终止（回滚）方法
 * @param {Transaction *} txn 需要回滚的事务
 * @param {LogManager} *log_manager 日志管理器指针
 */
void TransactionManager::abort(Transaction * txn, LogManager *log_manager) {
    // Todo:
    // 1. 回滚所有写操作
    // 2. 释放所有锁
    // 3. 清空事务相关资源，eg.锁集
    // 4. 把事务日志刷入磁盘中
    // 5. 更新事务状态
    std::scoped_lock lock{ latch_ };
    auto write_set=txn->get_write_set();
    while (!write_set->empty())
    {
        auto write=write_set->back();
        //得到操作记录的rid
        auto write_rid=write->GetRid();
        //得到操作的表
        auto write_table=write->GetTableName();
        //得到操作的表的元数据
        auto write_tablemeta=sm_manager_->db_.get_table(write_table);
        if(write->GetWriteType()==WType::DELETE_TUPLE){
            bool is_ok=true;
            //得到删除的记录
            auto write_rec=write->GetRecord();
            //一致性检查
            for(size_t i = 0; i < write_tablemeta.indexes.size(); ++i) {
                auto& index = write_tablemeta.indexes[i];
                auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(write_table, index.cols)).get();
                char* key = new char[index.col_tot_len];
                int offset = 0;
                for(size_t j = 0; j < index.col_num; ++j) {
                    memcpy(key + offset, write_rec.data + index.cols[j].offset, index.cols[j].len);
                    offset += index.cols[j].len;
                }
                std::vector<Rid> result;
                if(ih->get_value(key, &result, txn)){
                    delete[] key;
                    is_ok=false;
                    break; 
                }
                delete[] key;
            }
            if(is_ok==false){
                write_set->pop_back();
                delete write;
                continue; 
            }
            //写插入日志
            InsertLogRecord* insert_log=new InsertLogRecord(txn->get_transaction_id(),write_rec,write_rid,write_table);
            insert_log->prev_lsn_=txn->get_prev_lsn();
            log_manager->add_log_to_buffer(insert_log);
            txn->set_prev_lsn(insert_log->lsn_);
            delete insert_log;
            //把删除的记录加回去
            sm_manager_->fhs_.at(write_table)->insert_record(write_rid,write_rec.data);
            //添加索引
            for(size_t i = 0; i < write_tablemeta.indexes.size(); ++i){
                auto& index = write_tablemeta.indexes[i];
                auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(write_table, index.cols)).get();
                char* key=new char[index.col_tot_len];
                int off=0;
                for(int i=0;i<index.col_num;i++){
                    memcpy(key+off,write_rec.data+index.cols[i].offset,index.cols[i].len);
                    off=off+index.cols[i].offset;
                }
                ih->insert_entry(key, write_rid, nullptr);
                delete[] key;
            }
            

        }else if(write->GetWriteType()==WType::INSERT_TUPLE){
            //先删除索引
            auto insert_rec=sm_manager_->fhs_.at(write_table)->get_record(write_rid,nullptr);
            for(size_t i = 0; i < write_tablemeta.indexes.size(); ++i){
                auto& index = write_tablemeta.indexes[i];
                auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(write_table, index.cols)).get();
                char* key=new char[index.col_tot_len];
                int off=0;
                for(int i=0;i<index.col_num;i++){
                    memcpy(key+off,insert_rec->data+index.cols[i].offset,index.cols[i].len);
                    off=off+index.cols[i].offset;
                }
                ih->delete_entry(key,nullptr);
                delete[] key;
            }
            //写删除日志
            DeleteLogRecord* delete_log=new DeleteLogRecord(txn->get_transaction_id(),*insert_rec,write_rid,write_table);
            delete_log->prev_lsn_=txn->get_prev_lsn();
            log_manager->add_log_to_buffer(delete_log);
            txn->set_prev_lsn(delete_log->lsn_);
            delete delete_log;
             //把插入的记录删除,这个里面的事务指针有什么用？
            sm_manager_->fhs_.at(write_table)->delete_record(write_rid,nullptr);
            
        }else if(write->GetWriteType()==WType::UPDATE_TUPLE){
            auto prev_rec=write->GetRecord();
            auto update_rec=sm_manager_->fhs_.at(write_table)->get_record(write_rid,nullptr);
            //一致性检查
            bool is_ok=true;
            for(size_t i = 0; i < write_tablemeta.indexes.size(); ++i) {
                auto& index = write_tablemeta.indexes[i];
                auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(write_table, index.cols)).get();
                char* key = new char[index.col_tot_len];
                int offset = 0;
                for(size_t j = 0; j < index.col_num; ++j) {
                    memcpy(key + offset, prev_rec.data + index.cols[j].offset, index.cols[j].len);
                    offset += index.cols[j].len;
                }
                std::vector<Rid> result;
                if(ih->get_value(key, &result, txn)){
                    delete[] key;
                    is_ok=false;
                    break;
                }
                delete[] key;
            }
            if(is_ok==false){
                write_set->pop_back();
                delete write;
                continue; 
            }
            //写更新日志
            UpdateLogRecord* update_log=new UpdateLogRecord(txn->get_transaction_id(),prev_rec,*(update_rec.get()),write_rid,write_table);
            update_log->prev_lsn_=txn->get_prev_lsn();
            log_manager->add_log_to_buffer(update_log);
            txn->set_prev_lsn(update_log->lsn_);
            delete update_log;
            //先删除更新之后的索引
            for(size_t i = 0; i < write_tablemeta.indexes.size(); ++i){
                auto& index = write_tablemeta.indexes[i];
                auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(write_table, index.cols)).get();
                char* key=new char[index.col_tot_len];
                int off=0;
                for(int i=0;i<index.col_num;i++){
                    memcpy(key+off,update_rec->data+index.cols[i].offset,index.cols[i].len);
                    off=off+index.cols[i].offset;
                }
                char* new_key=new char[index.col_tot_len];
                off=0;
                for(int i=0;i<index.col_num;i++){
                    memcpy(new_key+off,prev_rec.data+index.cols[i].offset,index.cols[i].len);
                    off=off+index.cols[i].offset;
                }
                ih->delete_entry(key,nullptr);
                ih->insert_entry(new_key, write_rid, nullptr);
                delete[] key;
                delete[] new_key;
            }
            //然后将记录更新回去,这里的事务指针是？
            sm_manager_->fhs_.at(write_table)->update_record(write_rid,prev_rec.data,nullptr);
            //加索引
            // for(size_t i = 0; i < write_tablemeta.indexes.size(); ++i){
            //     auto& index = write_tablemeta.indexes[i];
            //     auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(write_table, index.cols)).get();
                
            //     int off=0;
                
                
            // }
            
        }    
        write_set->pop_back();
        delete write;    
    }
    //清楚write_set
    write_set->clear();
    // 2. 释放所有锁
    for(auto &lock:*(txn->get_lock_set())){
        lock_manager_->unlock(txn,lock);
    }
    // 3. 释放事务相关资源，eg.锁集
    txn->get_lock_set()->clear();
    // 4. 把事务日志刷入磁盘中
    AbortLogRecord* abort_log=new AbortLogRecord(txn->get_transaction_id());
    abort_log->prev_lsn_=txn->get_prev_lsn();
    log_manager->add_log_to_buffer(abort_log);
    log_manager->flush_log_to_disk();
    txn->set_prev_lsn(abort_log->lsn_);
    
    // 5. 更新事务状态
    txn->set_state(TransactionState::ABORTED);
    delete abort_log;
}